/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*    
      mflow.c  - mean flow module to turbulence model 
*/

#include <math.h>
#include <stdio.h>

#include "mflow.h"

const double fcoriolis = 1.0e-4;

switch_t mflow_flag=0;

double ustar;
double vstar;

double wave_forcx=0.0;
double wave_forcy=0.0;
double mFx=0.0;
double mFy=0.0;

/* Mean flow field variables */

field1D *UU;
field1D *VV;
field1D *UT1;
field1D *UT2;
field1D *VT1;
field1D *VT2;
field1D *KUV;        //  Sm*l*Q  - different from TKE eddy diffusion
field1D *UW;         //  <u'w'>  (m^2/s^2)     
field1D *VW;         //  <v'w'>  (m^2/s^2)

/* ------------ Functions --------------  */

/* --- Initializing mean flow functions -- */

void init_mflow_to_log(field1D *V, double vstar, double z0b)
{
  int N= V->N;
  int i;
  double z;

  for (i=0;i<N;i++) {
    z = i*dz + z0b;
    DR1(V,i) = vstar*log(z/z0b)/0.4;
  }

}


void geostrophic_ic(field1D *U, field1D *V, double Fx, double Fy)
{

  int N= V->N;
  int i;
  const double zmax=5;    // limit of initial bottom bbl in meters
  int imax;               // indexed limit of initial bbl
  double Vg, Ug;
  double fimax;

  Vg = -Fx/fcoriolis;
  Ug = Fy/fcoriolis;

  fimax = idz*zmax;
  imax = (int ) fimax;

  //  printf("Ug=%lf, Vg=%lf, zmax=%lf, idz=%lf, imax=%d\n",Ug,Vg,zmax,idz,imax);

  for (i=1;i<imax;i++) {
    DR1(V,i) = Vg * (i/fimax);
    DR1(U,i) = Ug * (i/fimax);
  }

  for (i=imax;i<N;i++) {
    DR1(V,i) = Vg;
    DR1(U,i) = Ug;
  }

}


void file_load_mflow_ic(field1D *U, field1D *V, char *ufile, char *vfile)
{
  int rval;

  rval = load_field1D_ascii(ufile,U);
  if (rval == 0)
    error_message("U file reading error");

  rval = load_field1D_ascii(vfile,V);
  if (rval == 0)
    error_message("V file reading error");

  /* check to make sure bottom bc is valid */

  if (DR1(U,0) != 0.0) {
    fprintf(stderr,"*** Warning, loaded U ic file doesn't have a no flow bed bc.\n");
    DR1(U,0) = 0.0;
  }

  if (DR1(V,0) != 0.0) {
    fprintf(stderr,"*** Warning, loaded V ic file doesn't have a no flow bed bc.\n");
    DR1(V,0) = 0.0;
  }

}

void allocate_mflow_variables(int nz, mflowinfo *M)
{

  mflow_flag = M->mflow_flag ;  // set whether mean flow stuff is on or off
  ustar = M->ustar;             // set the x - wind stress
  vstar = M->vstar;             // set the y - wind stress

  mFx = M->mFx;
  mFy = M->mFy;

  UU  = allocate_field1D(nz);
  VV  = allocate_field1D(nz);
  UT1 = allocate_field1D(nz-2);
  UT2 = allocate_field1D(nz-2);
  VT1 = allocate_field1D(nz-2);
  VT2 = allocate_field1D(nz-2);
  KUV = allocate_field1D(nz-1);
  UW = allocate_field1D(nz-1);
  VW = allocate_field1D(nz-1);

  set_field1D_to_value(UU,0.0);
  set_field1D_to_value(VV,0.0);

  //  init_mflow_to_log(VV,0.0154,0.005);

  set_field1D_to_value(UT1,0.0);
  set_field1D_to_value(UT2,0.0);

  set_field1D_to_value(VT1,0.0);
  set_field1D_to_value(VT2,0.0);

#ifdef ROTATION 
 #ifdef GEOSTROPHIC_IC

  geostrophic_ic(UU,VV,mFx,mFy);

 #endif  /* GEOSTROPHIC_IC */
 #ifdef SHAW_IC

  file_load_mflow_ic(UU,VV,"mUg.dat","mVg.dat");

 #endif  /* SHAW_IC */

#endif /* ROTATION */
}


void deallocate_mflow_variables()
{

  deallocate_field1D(UU);
  deallocate_field1D(VV);
  deallocate_field1D(UT1);
  deallocate_field1D(UT2);
  deallocate_field1D(VT1);
  deallocate_field1D(VT2);
  deallocate_field1D(KUV);
  deallocate_field1D(UW);
  deallocate_field1D(VW);

}






void rhs_implicit_mflow_eq(field1D *UT, field1D *VT, const field1D *U, const field1D *V)
{

  const double fcoriolis = 1e-4;

  int i;
  int N = U->N;
  double *utd = UT->data;
  double *vtd = VT->data;
  double *ud = U->data+1;
  double *vd = V->data+1;

  double *uwd = UW->data;
  double *vwd = VW->data;

  double vv, uu, vp, up;
  double urot=0.0, vrot=0.0;
  double dudt, dvdt;

  /* do the first grid minus flux before looping */


  for (i=1;i<N-1;i++) {

    vv = *vd++;   
    uu = *ud++;

    /* Here rotation is taken care of */

#ifdef ROTATION
    urot = fcoriolis * vv;   // check the signs here
    vrot = fcoriolis * uu;   // and here too
#endif /* ROTATION */

    dudt = wave_forcx + mFx + urot;
    dvdt = wave_forcy + mFy - vrot;

    *utd++ = dudt;
    *vtd++ = dvdt;
  }
}




