/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* q2l.c --- routines for the MY2.5 TKE dynamics */

#include <math.h>

#include "defs.h"
#include "tke.h"


extern field1D *LLZ;

const double Sqq = 0.2;   // units ????? also next 3??
const double Smm = 0.39;   /* Momentum mixing coef. in MY model */
double l;
const double Btt = 16.6;   /* dissipation const. for MY models */


/* ----------------------------------------------------------------------- */
/* ---Below is q^2 l equation functions ---------------------------------- */
/* ----------------------------------------------------------------------- */

void rhs_q2l_eq(field1D *BT, field1D *Q2LT, const field1D *B, const field1D *Q2L, 
                     const field1D *DISS, const field1D *PROD, const field1D *Kv)
{

  int i, N = DISS->N;

  static const double cL1 = 0.9;      // cL1 in q^2l equation in GOTM or E1 in MY82
  static const double E2 = 1.33;      // E2 in GOTM & MY82

  double *bd = B->data+1;
  double *q2ld = Q2L->data+1;
  double *dissd = DISS->data+1;
  double *q2ltd = Q2LT->data;
  double *btd = BT->data;
  double *kvd = Kv->data;
  //  double *kvm = Kv->data;
  double *prodd = PROD->data+1;
  double *diffd = TKEDIFF->data+1;

  double Ftke_p, Ftke_m, Fq2l_p, Fq2l_m;
  double bb, bp, q2l, q2lp, diss;
  double vgrad, shear_production, diffusion, buoyancy_flux;
  double kv,kvm; 
  double kl_diff, kl_prod, kl_diss, ll, ratio;
  //  double buoyancy_flux = 0.0;
  double dbdt, dq2ldt;


  /* first do the diffusion at the first grid cell */

  kvm = *kvd++;
  q2l = *q2ld++;

  bb = *bd++;

  Fq2l_m = kvm * (q2l - DR1(Q2L,0));
  Ftke_m = kvm * (bb - DR1(B,0));

  for (i=1;i<N-1;i++) {

  /* The tke diffusion */
  /* added in the variable length scale stuff */

    bp = *bd++;
    q2lp = *q2ld++;
    diss = *dissd++;
    kv = *kvd++;

    Ftke_p  = kv * (bp - bb);

    diffusion = idz2*(Ftke_p - Ftke_m);
    *diffd++ = diffusion;
    Ftke_m = Ftke_p;

    dbdt = diffusion - diss;

    vgrad = 0.25*idz2*(SQR(DR1(UU,i+1)-DR1(UU,i-1)) + SQR(DR1(VV,i+1)-DR1(VV,i-1)));
    shear_production = 0.5*(kvm+kv) * vgrad;   /* This is NOW fixed */
    kvm = kv;

    *prodd++ = shear_production;
    dbdt += shear_production;

    /* next do buoyancy flux */

#ifdef BUOYANCY_FLUX
    buoyancy_flux = DR1(BFLUX,i);
    dbdt += buoyancy_flux;
#endif /* BUOYANCY_FLUX */

#ifdef SEDIMENT_BUOYANCY_FLUX
    //    dbdt += DR1(SBFLUX,i);
#endif /* SEDIMENT_BUOYANCY_FLUX */

    *btd++ = dbdt;

    /* --------------------------------------- */
    /* --- Next do the q^2 l equation -----  */
    /* --------------------------------------- */

  /* The dissipation diffusion - also done slowly */
  /* added in the variable length scale stuff */

    Fq2l_p  = kv * (q2lp - q2l);    /* the q^2*l mixing coef is same as q^2/s, ie Sm=Sl */

    kl_diff = idz2*(Fq2l_p-Fq2l_m);
    Fq2l_m = Fq2l_p;

    kl_prod = cL1 * shear_production;

    ll = DR1(LL,i);
    ratio = ll/DR1(LLZ,i);
    kl_diss = (1+ E2*SQR(ratio))*diss;


    dq2ldt = kl_diff + ll*(kl_prod - kl_diss);

    *q2ltd++ = dq2ldt;

    q2l = q2lp;
    bb = bp;
  }

  /* This is for the plunging wave tke source */

  if ((wave_breaking ==ON) && (turb_source == PLUNGING)) {
    btd = BT->data;
    for (i=1;i<N-1;i++) 
      *btd++ = DR1(TKE_SOURCE,i);
  }


}

/* Boundary Condition Function got Mellor Yamada q^2 l model  */

void update_q2l_bc(field1D *BB, field1D *Q2L, field1D *Kv)
{
  int N = BB->N;

  double kvtop;
  double vgrad, prod, bb, qq;
  double ll02 = SQR(DR1(LL,0));
  double ll = DR1(LL,0);
  static double Smtrue = 0.39;

#ifdef NO_FLUX_TKE_BOTTOM_BC
  DR1(BB,0) = DR1(BB,1);
  DR1(Q2L,0) = DR1(Q2L,1);
#endif

#ifdef NO_SLIP_TKE_BOTTOM_BC
  DR1(BB,0) = 0.0;
  DR1(Q2L,0) = 0.0;
#endif

#ifdef MY_TKE_BOTTOM_BC    /* shear prod = dissipation */
  vgrad = idz2*(SQR(DR1(UU,1)-DR1(UU,0)) + SQR(DR1(VV,1)-DR1(VV,0)));
  bb = 0.5*Smtrue * ll02 * Btt * vgrad;
  DR1(BB,0) = bb;
  qq = sqrt(2.0*bb);
  prod = Smtrue * ll * qq * vgrad;
  DR1(Q2L,0) = 2.0*bb*ll;   // MY82 say Q2L=0 at z=0 -> wierd
  DR1(PROD,0) = prod;
  DR1(PROD,N-1) = DR1(PROD,N-2);    // this is the surface condition!!  ACK
#endif

  /* This is for the top BC */

  kvtop = DR1(Kv,N-2);
  DR1(BB,N-1) = DR1(BB,N-2) + wave_tke_surface_flux*dz/kvtop;

  DR1(Q2L,N-1) = 2.0*DR1(BB,N-1)*DR1(LL,N-1);   // here set the surface condition 


  /* this is for the surface stuff - isn't important dynamically but for 
     later analysis we include this in */

  DR1(PROD,N-1) = DR1(PROD,N-2);

}




void calculate_q2l_stats(field1D *Q, field1D *DISS, field1D *Kv, field1D *KUv, const field1D *B, const field1D *Q2L)
{

  int i;
  int N = B->N;
  double tmp;
  double *bd = B->data;
  double *qd = Q->data;
  double *q2ld = Q2L->data+1;
  double *dissd = DISS->data+1;
  double *kvd = Kv->data;
  double *kuvd = KUv->data;
  double *lld = LL->data+1;

  double bb, q2l, qq, ll, eps;

  /* Here we do i=0 */

  bb = *bd++;
  qq = sqrt((2.0*bb));
  *qd++ = qq;

  /* Now loop over the remaining */

  for (i=1;i<N-1;i++) {
    bb = *bd++;
    qq = sqrt((2.0*bb));
    q2l = *q2ld++;
    //    if (bb<1e-10)
    //      bb=1e-10;
    ll = q2l/(2.0*bb);
    eps = 2.0*qq*bb/(Btt*ll);

#ifdef DEBUG
    if (i<1)
      printf("bb=%g, qq=%g, q2l=%g, ll=%g, eps=%g\n",bb,qq,q2l,ll,eps);
#endif /* DEBUG */

    *qd++ = qq;
    *lld++ = ll;
    *dissd++ = eps;

  }

  /* This is the top i=N-1 */

  bb = *bd++;
  qq = sqrt((2.0*bb));
  *qd++ = qq;

  /* for pretty sake */

  DR1(DISS,0) = DR1(DISS,1);
  DR1(DISS,N-1) = DR1(DISS,N-2);

  /* Here calculate the vertical eddy viscosities for momentum & tke */

  for (i=0;i<N-1;i++) {
    tmp = 0.25*(DR1(Q,i)+DR1(Q,i+1))*(DR1(LL,i)+DR1(LL,i+1));
    *kvd++ = Sqq*tmp + 1.2e-6;
    *kuvd++ = Smm*tmp + 1.2e-6;
    }

}


/* -----------------------------------------------------------------
   ----------------- q^2 l updating functions ----------------------
   -----------------------------------------------------------------  
   These are to be deprecated..... new semi-implicit functiosn must be written

*/

void update_q2l_AB2(field1D *B, const field1D *BT1, const field1D *BT2, 
                    field1D *Q2L, const field1D *Q2LT1, const field1D *Q2LT2, const double dt)
{

  const double *bt1d = BT1->data;
  const double *bt2d = BT2->data;
  const double *q2lt1d = Q2LT1->data;
  const double *q2lt2d = Q2LT2->data;

  double *bd = B->data+1;
  double *q2ld = Q2L->data+1;
  double hdt = 0.5*dt;

  int i, N;

  N = B->N;

  /* loop over N-2 the B and dissipation */

  for (i=1;i<N-1;i++) {
      *bd++ += hdt*(3.0* (*bt1d++) - *bt2d++);
      *q2ld++ += hdt*(3.0* (*q2lt1d++) - *q2lt2d++);
    }

  bd = B->data+1;
  q2ld = Q2L->data+1;

  for (i=1;i<N-1;i++) {
    if (*bd<0) 
      *bd = 1e-6;
    bd++;
    if (*q2ld<(1e-8)) 
      *q2ld = 1e-8;
    q2ld++;
  }

}


void update_q2l_euler(field1D *B, const field1D *BT1, 
                      field1D *Q2L, const field1D *Q2LT1, const double dt)
{

  const double *bt1d = BT1->data;
  const double *q2lt1d = Q2LT1->data;

  double *bd = B->data+1;
  double *q2ld = Q2L->data+1;

  int i, N;

  N = B->N;

  /* loop over all the B or Q points */

  for (i=1;i<N-1;i++) {
    *bd++ += dt* (*bt1d++);
    *q2ld++ += dt* (*q2lt1d++);
    }

  bd = B->data+1;
  q2ld = Q2L->data+1;

  for (i=1;i<N-1;i++) {
    if (*bd<0)
      *bd = 0.0;
    bd++;
    if (*q2ld<0)
      *q2ld = 0.0;
    q2ld++;
  }


}


