/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


/*    
      sediment.c  - sediment module to turbulence model 
*/

//#define SEDIMENT_DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <glib.h>

#include "sediment.h"
#include "tri.h"


switch_t sediment_flag = OFF;


/* Extern sediment field variables */

field1D *SS;
field1D *ST1;
field1D *ST2;
field1D *KSV;             //  Sm*l*Q, sediment eddy diffusivity - different from TKE eddy diffusion
field1D *SBFLUX;          // sediment buoyancy flux feedback into TKE eq

field1D *SFLUX_U;         // the x- sediment flux 
field1D *SFLUX_V;         // the y- sediment flux


/* Non-extern sediment variables */

double *rhss;        // these are for the implicit solver
double *dds;
double *uds;
double *lds;
double *works;

/* constant variables */

const double sgrav = 2.65;   // Trowbridge quotes this


double fall_velocity;
double D50;
double tau_b;         // this is the bed shear stress, u_*^2  (m^2/s^2)
double psi_critical;
double pickup_factor;
double psi;          // this is the shields parameter

bc_t   sediment_bed_bc;
double pickup;              // this is the bed sediment flux
double ref_concentration;   // this would be the bed reference C

/* Functions */

/* This returns the fall velocity for a particule in m/s
   The form in Fredsoe & Deigaard is used -> Cd = 36/Re + 1.4 
   which leads to a quadratic equation in ws
   1.4 w_s^2 + (36\nu/d) w_s  - 4gd(s-1)/3 = 0
   Note that s is specific gravity and for sand = 2.65  */

double sediment_fall_velocity(double D50)
{

  double a, b, c;
  double fac, ws;

  a = 1.4;
  b = 36*nu/D50;
  c = -4.0*grav*D50*(sgrav-1)/3;

  fac = b*b - 4.0*a*c;

  ws = (-b + sqrt(fac))/(2*a);

#ifdef SEDIMENT_DEBUG
  printf("** sediment.c: For D50=%.3g (m), Fall velocity ws = %.3g (m/s)\n",D50,ws);
#endif

  return ws;

}

/*
double shields_parameter(double ustar, double D50)
{

  double psi;

  psi = ustar*ustar/((sgrav-1)*grav*D50);

  return psi;

}
*/

/* Calculates RHS of dC/dt = ....  where C is sediment concentration */


void rhs_implicit_sediment_eq(field1D *ST, const field1D *S, const field1D *Kv)
{

  int i;
  int N = S->N;
  double *std = ST->data;
  double *sd = S->data+1;
  double *sdp = S->data+2;
  double *sdm = S->data;

  double *kv = Kv->data+1;

  double ss, sp;
  double t1, t2;
  double Fsp, Fsm;
  double sdiffusion, fall;
  double dsdt;

  Fsm = DR1(Kv,0)*(DR1(S,1)-DR1(S,0))*idz;

  for (i=1;i<N-3;i++) {

    ss = *sd++;
    sp = *sdp++;

    Fsp = (*kv++) * (sp - ss);
    sdiffusion = 0.5*idz2*(Fsp-Fsm);  // !! only implicit change (need 0.5)
    DR1(SBFLUX,i) = grav*0.5*(Fsp+Fsm)/rho0;  // is the /rho0 correct ???!!!! 
    Fsm = Fsp;                    // now up becomes minus!

    /* handle fall velocity here !!!!! */
    /* 3rd order upwind */

    //fall = -fall_velocity*idz*(2.0*DR1(SS,i-1) + 3.0*DR1(SS,i)-6.0*DR1(SS,i+1)+DR1(SS,i+2))/6.0;
    t1 = sp*(1-sp);
    t2 = ss*(1-ss);
    //fall = fall_velocity*idz*(t1-t2);
    fall = fall_velocity*idz*(sp-ss);
    dsdt = fall + sdiffusion;

    *std++ = dsdt;
  }

  /* .... i=N-3 ...... */

  i=N-3;
 
  ss = *sd++;
  sp = *sdp++;
  Fsp = (*kv++) * (sp - ss);
  sdiffusion = 0.5*idz2*(Fsp-Fsm);  /* !! only implicit change (need 0.5) */
  DR1(SBFLUX,i) = grav*0.5*(Fsp+Fsm)/rho0;  /* need buoyancy flux */
  Fsm = Fsp;

  /* handle fall velocity here !!!!! */
  /* 3rd order upwind and apply BC as if surface SS is ZERO */

  //fall = -fall_velocity*idz*(2.0*DR1(SS,i-1) + 3.0*DR1(SS,i)-6.0*DR1(SS,i+1))/6.0;
  t1 = sp*(1-sp);
  t2 = ss*(1-ss);
  //fall = fall_velocity*idz*(t1 - t2);
  fall = fall_velocity*idz*(sp - ss);
  dsdt = fall + sdiffusion;

  *std++ = dsdt;

  /* ...... i=N-2 ....... */

  i=N-2;
  ss = *sd++;
  sp = *sdp++;

  Fsp = (*kv++) * (sp - ss);
  sdiffusion = idz2*(Fsp-Fsm);
  DR1(SBFLUX,i) = grav*0.5*(Fsp+Fsm)/rho0;  /* is the /rho0 correct ???!!!! */

  /* handle fall velocity here !!!!! */
  /* 1st order upwind - and since we apply zero falling flux it's 
     as if SS(N-1)=0 for this                         */
  //  t1 = sp*(1-sp);
  t2 = ss*(1-ss);
  //fall = fall_velocity*idz*(- t2);
  fall = fall_velocity*idz*(-ss);

  dsdt = fall + sdiffusion;
  *std++ = dsdt;

}





void implicit_solve_sediment_euler(field1D *S, const field1D *ST,
                                   const field1D *Kv, double dt)
{

  int i, ok;
  int N = S->N;

  double *sd = S->data+1;

  const double *std = ST->data;
  const double *kvd = Kv->data;

  double ss, sp;
  double kv, kvm, total;
  double Fp, Fm, tmp;
  double sdiffusion;

  double hdt = 0.5*dt;

  /* First calculate the RHS stuff */
  /* no diffusion at step n need be calculated! */

  for (i=1;i<N-1;i++) {

    ss = *sd++;
    rhss[i-1] = ss + dt*(*std++);

  }   /* for (i=0.... */

  /* Now the tridiagonal matrix is set up */
  /* i = 0 -> here we apply no-slip bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;


  uds[0] = -hdt*kv*idz2;
  lds[0] = 0.0;
  kvm = kv;

  switch (sediment_bed_bc) {
  case FLUX: total = kvm;
     rhss[0] += idz*hdt*pickup;
     break;
  case FIXED: total = kvm+kv;
     rhss[0] += hdt*kv*idz2*ref_concentration;   // this is for fixed bc
  }

  dds[0] = 1 + hdt*idz2*(total);   // bed is no-flux bc.

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    dds[i] = 1 + hdt*idz2*(kvm+kv);
    uds[i] = -hdt*kv*idz2;
    lds[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  lds[i] = -hdt*kvm*idz2;
  uds[i] = 0.0;
  dds[i] = 1.0 + hdt*idz2*(kvm);

  /* Now do the implicit solving */

  sd = S->data+1;   // reset the sediment pointer

  ok = tridiagonal(dds, lds, uds, rhss, sd, works, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for density failed!");

  /* Next is sediment boundary conditions */
  /* Bottom BC is no density flux near bed */

  DR1(S,0) = DR1(S,1);

  /* Surface BC */

  switch(sediment_bed_bc) {
  case FIXED:  DR1(S,0) = ref_concentration;
               break;
  case FLUX:   DR1(S,N-1) = DR1(S,N-2) + pickup*dz/kv;
  }


}


void implicit_solve_sedimentAB2(field1D *S, const field1D *ST1, const field1D *ST2,
                                   const field1D *Kv, double dt)
{

  int i, ok;
  int N = S->N;

  double *sd = S->data+1;

  const double *std1 = ST1->data;
  const double *std2 = ST2->data;
  const double *kvd = Kv->data;

  double ss, sp;
  double kv, kvm, total;
  double Fp, Fm, tmp;
  double sdiffusion;

  double hdt = 0.5*dt;

  /* First calculate the RHS stuff */
  /* no diffusion at step n need be calculated! */

  for (i=1;i<N-1;i++) {

    ss = *sd++;
    rhss[i-1] = ss + hdt*(3.0*(*std1++) - (*std2++));

  }   /* for (i=0.... */

  /* Now the tridiagonal matrix is set up */
  /* i = 0 -> here we apply no-slip bottom bc */

  kvd = Kv->data;

  kvm = *kvd++;
  kv = *kvd++;


  uds[0] = -hdt*kv*idz2;
  lds[0] = 0.0;
  kvm = kv;

  switch (sediment_bed_bc) {
  case FLUX: total = kvm;
     rhss[0] += idz*hdt*pickup;
     break;
  case FIXED: total = kvm+kv;
     rhss[0] += hdt*kv*idz2*ref_concentration;   // this is for fixed bc
  }

  dds[0] = 1 + hdt*idz2*(total);   // bed is no-flux bc.

  /* i=1 to < N-3 */

  for (i=1;i<N-3;i++) {
    kv = *kvd++;
    dds[i] = 1 + hdt*idz2*(kvm+kv);
    uds[i] = -hdt*kv*idz2;
    lds[i] = -hdt*kvm*idz2;
    kvm = kv;
  }

  /* i = N-3, here deal with top flux BC */

  i = N-3;
  kv = *kvd++;
  lds[i] = -hdt*kvm*idz2;
  uds[i] = 0.0;
  dds[i] = 1.0 + hdt*idz2*(kvm);

  /* Now do the implicit solving */

  sd = S->data+1;   // reset the sediment pointer

  ok = tridiagonal(dds, lds, uds, rhss, sd, works, N-2);
  if (ok != 0)
    error_message("** Critical Error:  Tridiagonal call for density failed!");

  /* Next is sediment boundary conditions */
  /* Bottom BC is no density flux near bed */

  DR1(S,0) = DR1(S,1);

  /* Surface BC */

  switch(sediment_bed_bc) {
  case FIXED:  DR1(S,0) = ref_concentration;
               break;
  case FLUX:   DR1(S,N-1) = DR1(S,N-2) + pickup*dz/kv;
  }


}


/* ...... */

static double denom1bc;

void calculate_sediment_stats(field1D *Kv)
{
  int i, N = SBFLUX->N;
  double tau_x, tau_y;
  double Fsp, Fsm;
  double kvtop = DR1(Kv,nz-2);
  double psi_ratio;

  double ss, *sd = SS->data;
  double *sp = SS->data+1;
  double *kv = Kv->data;

  /* calculate the bed shear stress */

  tau_x = DR1(KUV,1)*idz*(DR1(UU,2)-DR1(UU,1));
  tau_y = DR1(KUV,1)*idz*(DR1(VV,2)-DR1(VV,1));
  tau_b = sqrt(tau_x*tau_x + tau_y*tau_y);

  /* Here we use the same mixing coef as for turbulence */

  copy_field1D(Kv,KQV);  

  /* This pickup function is based on that in Neilsens book! */
  /* Bottom Boundary Condition */
  // 1. calculate sheilds stress, need bottom stress first
  // 2. 

  /* calculate the bottom stress & shields parameter */

  psi = tau_b/denom1bc;  // demon1bc = (sgrav-1)*grav*D50; (see further down)

  if (psi > psi_critical) {
    psi_ratio = (psi-psi_critical)/psi_critical;
    pickup = psi_ratio*sqrt(psi_ratio)*pickup_factor;
  }
  else {
    pickup = 0.0;
  }

#ifdef SEDIMENT_DEBUG
  printf("** sediment.c: psi = %g,  psi_c = 0.06, pickup=%g, pickup_factor=%g\n",psi,pickup,pickup_factor);
#endif

}




void allocate_sediment_variables(int nz, sedimentinfo *SED)   //int flag, double grain_sizeD50)
{
  double tmp;
 
  sediment_flag = SED->sediment_flag;  // set whether sediment module is on or off
  
  D50 = SED->D50;
  fall_velocity = sediment_fall_velocity(D50);

  sediment_bed_bc = FLUX;   //!!!! this is now hard wired!
  /* allocate the memory to be used */

  SS  = allocate_field1D(nz);
  ST1 = allocate_field1D(nz-2);
  ST2 = allocate_field1D(nz-2);
  KSV = allocate_field1D(nz-1);
  SBFLUX = allocate_field1D(nz);
  SFLUX_U = allocate_field1D(nz);  // x-sediment flux
  SFLUX_V = allocate_field1D(nz);  // y-sediment flux

  set_field1D_to_value(SS,0.0);
  set_field1D_to_value(ST1,0.0);
  set_field1D_to_value(ST2,0.0);
  set_field1D_to_value(KSV,0.0);
  set_field1D_to_value(SBFLUX,0.0);
  set_field1D_to_value(SFLUX_U,0.0);
  set_field1D_to_value(SFLUX_V,0.0);

  /* implicit solver terms */

  dds = (double * ) g_malloc(sizeof(double)*(nz-2));
  uds = (double * ) g_malloc(sizeof(double)*(nz-2));
  lds = (double * ) g_malloc(sizeof(double)*(nz-2));
  rhss = (double * ) g_malloc(sizeof(double)*(nz-2));
  works = (double * ) g_malloc(sizeof(double)*(nz-2));

  /* terms for the pickup function */

  psi_critical = 0.056;     // this is from Madsen notes, FD & Peter N. book (about 0.06)
  tmp = (sgrav-1)*grav*D50*D50*D50/(nu*nu);
  pickup_factor = 0.00033*pow(tmp,0.1) * sqrt((sgrav-1)*grav*D50);   // check which one is correct!!!
  denom1bc = (sgrav-1)*grav*D50;


}


void deallocate_sediment_variables()
{

  deallocate_field1D(SS);
  deallocate_field1D(ST1);
  deallocate_field1D(ST2);
  deallocate_field1D(KSV);
  deallocate_field1D(SBFLUX);
  deallocate_field1D(SFLUX_U);
  deallocate_field1D(SFLUX_V);

  g_free(dds);
  g_free(uds);
  g_free(lds);
  g_free(works);
  g_free(rhss);

}


void calculate_sediment_fluxes(field1D *SFX, field1D *SFY)
{

  int i;
  int N = SFX->N;
  double *sfxd = SFX->data;
  double *sfyd = SFY->data;
  double *ssd = SS->data;
  double *ud = UU->data;
  double *vd = VV->data;
  double ss;
  
  for (i=0;i<N;i++) {
    ss = *ssd++;
    *sfxd++ = ss * (*ud++);
    *sfyd++ = ss * (*vd++);
  }

}



/*
void rhs_sediment_eq(field1D *ST, const field1D *S, const field1D *Kv)
{

  int i;
  int N = S->N;
  double *std = ST->data;
  double *sd = S->data+1;
  double *sdp = S->data+2;
  double *sdm = S->data;

  double *kv = Kv->data+1;

  double ss, sp;
  double t1, t2;
  double Fsp, Fsm;
  double sdiffusion, fall;
  double dsdt;

  Fsm = DR1(Kv,0)*(DR1(S,1)-DR1(S,0))*idz;

  for (i=1;i<N-3;i++) {

    ss = *sd++;
    sp = *sdp++;
    Fsp = (*kv++) * (sp - ss);
    sdiffusion = idz2*(Fsp-Fsm);
    DR1(SBFLUX,i) = grav*0.5*(Fsp+Fsm)/rho0;  // is the /rho0 correct ???!!!! 
    Fsm = Fsp;                    // now up becomes minus!



    //fall = -fall_velocity*idz*(2.0*DR1(SS,i-1) + 3.0*DR1(SS,i)-6.0*DR1(SS,i+1)+DR1(SS,i+2))/6.0;
    t1 = sp*(1-sp);
    t2 = ss*(1-ss);
    //fall = fall_velocity*idz*(t1-t2);
    fall = fall_velocity*idz*(sp-ss);
    dsdt = fall + sdiffusion;

    *std++ = dsdt;
  }

  // .... i=N-3 ...... 

  i=N-3;
 
  ss = *sd++;
  sp = *sdp++;
  Fsp = (*kv++) * (sp - ss);
  sdiffusion = idz2*(Fsp-Fsm);
  DR1(SBFLUX,i) = grav*0.5*(Fsp+Fsm)/rho0;
  Fsm = Fsp;

// handle fall velocity here !!!!! 
  // 3rd order upwind and apply BC as if surface SS is ZERO 

  //fall = -fall_velocity*idz*(2.0*DR1(SS,i-1) + 3.0*DR1(SS,i)-6.0*DR1(SS,i+1))/6.0;
  t1 = sp*(1-sp);
  t2 = ss*(1-ss);
  //fall = fall_velocity*idz*(t1 - t2);
  fall = fall_velocity*idz*(sp - ss);
  dsdt = fall + sdiffusion;

  *std++ = dsdt;

  // ...... i=N-2 ....... 

  i=N-2;
  ss = *sd++;
  sp = *sdp++;

  Fsp = (*kv++) * (sp - ss);
  sdiffusion = idz2*(Fsp-Fsm);
  DR1(SBFLUX,i) = grav*0.5*(Fsp+Fsm)/rho0;  // is the /rho0 correct ???!!!! 

  // handle fall velocity here !!!!! 
  //  1st order upwind - and since we apply zero falling flux it's 
     as if SS(N-1)=0 for this                         
  //  t1 = sp*(1-sp);
  t2 = ss*(1-ss);
  //fall = fall_velocity*idz*(- t2);
  fall = fall_velocity*idz*(-ss);

  dsdt = fall + sdiffusion;
  *std++ = dsdt;

}

*/
