/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */


#include <stdio.h>

#include "tri.h"


/* ------ TRIDIAGONAL MATRIX SOLVER ROUTINES ------------- 


First is the real valued tridiagonal matrix solver routinues
takes the diagonal, and the upper and lower and returns the answer in x
AD, ADU, and ADL are all length N vectors.

The first element of ADL is zero and the last element of ADU is
zero.

on return:  

  0:  ok
 -1:  bad  (diag[0]=0.0)
  j:  location of zero pivot

*/

/* Called for tke and epsilon equations (and sediment) */

int  tridiagonal(double *ad,double *adl,double *adu,double *b,double *x,double *work,int n)
{

    int j;
    double bet;

    if (ad[0]==0.0) 
      return -1;
    
    x[0]=b[0]/(bet=ad[0]);

    for (j=1;j<n;j++) {
       work[j]=adu[j-1]/bet;
       bet = ad[j]-adl[j]*work[j];
       if (bet == 0.0) {
       	 printf("** bet=%g, ad[j]=%g, adl[j]=%g, work[j]=%g\n",bet,ad[j],adl[j],work[j]);
	 return j;
       }
       x[j]=(b[j]-adl[j]*x[j-1])/bet;
    }
    for (j=n-2;j>=0;j--) 
       x[j] -= work[j+1]*x[j+1];
    
    return 0;

}


/* This is for the mean flow */

int  tridiagonal_mflow(const double *ad, const double *adl, const double *adu,
                       double *bu, double *bv, double *xu, double *xv,
                       double *work,int n)
{

    int j;
    double bet;

    if (ad[0]==0.0) 
      return 1;
    
    bet = ad[0];
    xu[0]=bu[0]/bet;
    xv[0]=bv[0]/bet;

    for (j=1;j<n;j++) {
       work[j]=adu[j-1]/bet;
       bet = ad[j]-adl[j]*work[j];
       if (bet == 0.0) 
          return 2;
       xu[j]=(bu[j]-adl[j]*xu[j-1])/bet;
       xv[j]=(bv[j]-adl[j]*xv[j-1])/bet;
    }
    for (j=n-2;j>=0;j--) {
       xu[j] -= work[j+1]*xu[j+1];
       xv[j] -= work[j+1]*xv[j+1];
    }
    
    return 0;

}


/* Called for the bubble solver */

int  tridiagonal_bubble(const double *ad, const double *adl, const double *adu,
                        const field2D *BB, field2D *XX, double *work, double *work_bet)
{
  int i,j;

  int M = BB->M;
  int N = BB->N;

  double *b;
  double *x;

  if (ad[0]==0.0)
    return 1;
    
  work_bet[0] = ad[0];
  for (i=0;i<N;i++) {
    x = XX->row[i];
    b = BB->row[i];
    x[0]=b[0]/work_bet[0];
  }

  for (j=1;j<M;j++) {
    work[j]=adu[j-1]/work_bet[j-1];
    work_bet[j] = ad[j]-adl[j]*work[j];
    if (work_bet[j] == 0.0)
      return 2;
  }

  for (i=0;i<N;i++) {
    x = XX->row[i];
    b = BB->row[i];
    for (j=1;j<M;j++) {
      x[j]=(b[j]-adl[j]*x[j-1])/work_bet[j];
    }
    for (j=M-2;j>=0;j--) {
       x[j] -= work[j+1]*x[j+1];
    }
  }    

  return 0;


}



/* Called for the bubble solver */

int  tridiagonal_bubble2(const double *ad, const double *adl, const double *adu,
                        const field2D *BB, field2D *XX, double *work, double *work_bet)
{
  int i,j,k;

  int M = BB->M;
  int N = BB->N;

  double *b;
  double *x;

  if (ad[0]==0.0)
    return 1;
    
  work_bet[0] = ad[0];
  x = XX->data;
  b = BB->data;

  for (i=0;i<M;i++) {
    *x++ = *b++/work_bet[0];
  }

  for (j=1;j<N;j++) {
    work[j]=adu[j-1]/work_bet[j-1];
    work_bet[j] = ad[j]-adl[j]*work[j];
    if (work_bet[j] == 0.0)
      return 2;
  }

  x = XX->data;
  b = BB->data;

  for (i=0;i<M;i++) {
    for (j=1,k=M;j<N;j++) {
      x[k]=(b[k]-adl[j]*x[k-M])/work_bet[j];
      k += M;
    }
    for (j=N-2,k=M*(N-2);j>=0;j--) {
      x[k] -= work[j+1]*x[k+M];
      k -= M;
    }
    x++;
    b++;
  }    

  return 0;


}


