/*
 * Copyright (c) 2002 Falk Feddersen
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/* 
   wave.c 

   Here all the wave stuff is done.   Changing the U(z),
   determining if wave breaking is occuring, and the 
   delta function tke & bubble sources 

*/

#include <stdio.h>
#include <math.h>
#include "wave.h"
#include "mflow.h"
#include "bubble.h"
#include "wavemodel.h"

/* here are the defines */

/*  These two are never used and may be implemented one day???
//#define MONOCHROMATIC
// #define RANDOM
*/

#define CUBE(a)  ((a)*(a)*(a))

/* Here are the global variables for this file only */


double Hrms;          // rms wave height
double T;             // This is the wave period
double theta;         // this is the wave angle in radians
int    Tn;            // number of time steps per wave period
double freq;          // wave freq
double omega;         // radian freq
double wavenum;       // wave number
double group_v ;      // the group velocity
double phase_v;       // wave phase velocity
double mean_epsilon;  // wave averaged wave dissipation
double max_forcx;      // maximum pressure force 
double max_forcy;      // maximum pressure force 
double max_forc;

#ifdef SHAW_TIDAL_FORCING
 double max_tidal_forc;
 double omega_tide;
#endif /* SHAW_TIDAL_FORCING */


breaking_t turb_source;
switch_t wave_breaking;
double wave_tke_surface_flux=0.0;

switch_t wave_forcing;


const double gam = 0.4;
const double B = 1.0;

field1D *TKE_SOURCE;          // plunging tke source
field1D *BETA_SOURCE;         // plunging void fraction source 
field1D *BASE_TKE_SOURCE;
field1D *BASE_BETA_SOURCE;
field2D *NN_SOURCE;
field2D *BASE_NN_SOURCE;
field1D *WAVE_NN_SURFACE_FLUX;
field1D *FRAC_WD;




/* This has BPE==TKE source, BPE~exp(-z'/scale)  TKE~ z'*exp(z'/scale_h) */

void setup_tke_beta_sources(field1D *TKE_S, field1D *BETA_S, double scale_h)
{
  int i;
  int N = BETA_S->N;
  double tot_pe = 0.0;
  double tot_tke = 0.0;
  double z, z0, zp;

  const double *betad = BETA_S->data;
  const double *betadp = BETA_S->data+1;

  /* first setup the BS to exponentially decay */

  for (i=0;i<N-1;i++) {
    z = depth - i*dz;
    DR1(BETA_S,i) = exp(-z/scale_h);
    DR1(TKE_S,i) = z * exp(-z/scale_h);
  }

  /* calculate the BPE & TKE */

  for (i=0;i<N-1;i++) {
    z0 = i*dz;
    zp = z0+dz; 
    tot_pe += 0.5*(z0 * *betad++ + zp * *betadp++);
    tot_tke += 0.5*(DR1(TKE_S,i)+DR1(TKE_S,i+1));
  }    
  tot_pe *= grav*dz;
  tot_tke *=dz;

  /* now normalize so that the total bpe is 1.0 */

  for (i=0;i<N;i++) {
    DR1(BETA_S,i) /= tot_pe;
    DR1(TKE_S,i) /= tot_tke;

  }

#ifdef DEBUG   /* check that bpe == 1.0 */
  tot_pe = 0.0;
  tot_tke = 0.0;
  for (i=0;i<N-1;i++) {
    z0 = i*dz;
    zp = z0+dz; 
    tot_pe += 0.5*(z0 * DR1(BETA_S,i) + zp * DR1(BETA_S,i+1));
    tot_tke += 0.5*(DR1(TKE_S,i)+DR1(TKE_S,i+1));
  }    
  tot_pe *= grav*dz;
  tot_tke *= dz;
  fprintf(stdout,"** wave.c:setup_beta_source():  tot_tke_source = %lf, tot_bpe_source = %lf\n",tot_tke,tot_pe);
#endif /* DEBUG */


}

/* This has the TKE and BPE source ~ exp(-z'/scale_h) */

void setup_tke_beta_sources2(field1D *TS, field1D *BS, double scale_h)
{
  int i;
  int N = BS->N;
  double tot_pe = 0.0;
  double tot_tke = 0.0;
  double z, z0, zp;
  double zeta;

  const double *betad = BS->data;
  const double *betadp = BS->data+1;

  /* adjust scale height to match Hsig */

  scale_h *= sqrt(2.0);

  /* first setup the BS to exponentially decay */

  DR1(BS,0) = 1.0;
  DR1(TS,0) = 1.0;
  for (i=0;i<N-1;i++) {
    z = depth - (i-1)*dz;
    zeta = z/(scale_h/sqrt(2.0));   //SQR(z/scale_h);
    DR1(BS,i) = exp(-zeta);
    DR1(TS,i) = exp(-zeta);
  }

  /* calculate the BPE & TKE */

  for (i=0;i<N-1;i++) {
    z0 = i*dz;
    zp = z0+dz; 
    tot_pe += 0.5*(z0 * *betad++ + zp * *betadp++);
    tot_tke += 0.5*(DR1(TS,i)+DR1(TS,i+1));
  }    
  tot_pe *= grav*dz;
  tot_tke *=dz;

  /* now normalize so that the total bpe is 1.0 */

  for (i=0;i<N;i++) {
    DR1(BS,i) /= tot_pe;
    DR1(TS,i) /= tot_tke;

  }

#ifdef DEBUG   /* check that bpe == 1.0 */
  tot_pe = 0.0;
  tot_tke = 0.0;
  for (i=0;i<N-1;i++) {
    z0 = i*dz;
    zp = z0+dz; 
    tot_pe += 0.5*(z0 * DR1(BS,i) + zp * DR1(BS,i+1));
    tot_tke += 0.5*(DR1(TS,i)+DR1(TS,i+1));
  }    
  tot_pe *= grav*dz;
  tot_tke *= dz;
  fprintf(stdout,"** wave.c:setup_beta_source():  tot_tke_source = %lf, tot_bpe_source = %lf\n",tot_tke,tot_pe);
#endif /* DEBUG */


}



double setup_fracwd(int Tn, double mepsilon, double dt, double T)
{

  int i;
  int width=1; 
  int start;
  double fepsilon;
  double tot;

  /* first do a sanity check ! */

  if (Tn<1)
    error_message("** Attempt to use Tn < 0");

  width = Tn/20;   /* this is the normal time dependent stuff */
  start = Tn/2;

#ifdef STEADY_WAVE_FORCING 
  width = Tn;     // this is for steady tke sources 
  start = 0;
  printf("** kepsilon: Warning: Using steady source of tke!!\n");
#endif

  FRAC_WD = allocate_field1D(Tn);
  set_field1D_to_value(FRAC_WD,0.0);

  fepsilon = mepsilon*T/(width*dt);

  for (i=start;i<(start+width);i++) {
    DR1(FRAC_WD,i) = fepsilon;
  }

#ifdef DEBUG
  tot = 0.0;
  for (i=0;i<Tn;i++) 
    tot += DR1(FRAC_WD,i);
  tot *= dt/T;
  fprintf(stdout,"** wave.c:setup_fracwd():  width=%d, Tn=%d, (1/T)*Integral FRAC_WD dt= %lf\n",width,Tn,tot);
#endif /* DEBUG */

  return fepsilon;
}


/* Initialize all the global constants related to the waves */

void set_wave_info(waveinfo *W, double dt, mflowinfo *M)
{

  double fTn;
  double dsxydx;
  Hrms = W->Hrms;
  T = W->T;
  theta = W->theta;

  freq = 1.0/T;
  omega = 2*pi*freq;

  fTn = T/dt;
  Tn = (int ) fTn;

  wavenum = get_wavenum(omega,depth);
  group_v = get_cg(wavenum, depth);
  phase_v = omega/wavenum;

  W->wave_num = wavenum;
  W->group_v = group_v;
  W->phase_v = phase_v;

  /* Here are the wave forcing parameters */

  wave_forcing = W->wave_forcing;
  max_forc = 0.5*wavenum*grav*Hrms;
  max_forcx = max_forc * cos(theta);
  max_forcy = max_forc * sin(theta);
  W->U0 = max_forc/omega;             // this should be inviscid shallow wave velocity
  turb_source = W->turb_source;

#ifdef DEBUG
  printf("WAVE: Hrms = %lf, T=%lf, f=%lf, omega = %lf, depth=%lf, wavenum=%lf, Cg=%lf, C=%lf, theta=%lf\n",Hrms,T,freq,omega,depth,wavenum,group_v,phase_v,theta*180/pi);
  printf("WAVE: max_forc: = %g\n",max_forc);
#endif /* DEBUG */

  /* Here either pick TG83 specified wave diss or the user selected */

  if (W->wepsilon_specified == -1.0)
    mean_epsilon = wave_averaged_energy_dissipation(Hrms, depth, freq, gam, B);
  else
    mean_epsilon = W->wepsilon_specified;

  W->mepsilon = mean_epsilon;

  /* set up the mean forcing possibilities */

  dsxydx = mean_epsilon*sin(theta)/phase_v;
  W->dsxydx = dsxydx;
  W->Fy = dsxydx/depth;

  if (M->surfzone_flag == ON)   /* if the "surfzone" option is set */
    M->mFy = W->Fy;

#ifdef SHAW_TIDAL_FORCING
  max_tidal_forc = M->mFx;
  omega_tide = 2*pi/(12*3600);   // the frequency of a tide
  M->mFx = 0;
  M->mFy = 0;
#endif  /* SHAW_TIDAL_FORCING */


}


void init_wave_parameters(int nz, int nr, waveinfo *W, double scale_h, double dt, mflowinfo *MF, double z0s)
{

  static const double n0 = 1e+4;
  static const double a0 = 0.0013;  // 1.3 mm

  int i,j;
  int N;
  int M;

  double *nnsd, *nnsd1;
  double *rd;
  double *wnsfd;

  double aa;
  double tmp, betam, betat, beta, rr;
  double betaf; 
  double fepsilon;
  double scale_height;

  set_wave_info(W, dt, MF);  // init epsilon, U0, etc.

  /* allocate the memory */

  TKE_SOURCE = allocate_field1D(nz);
  BETA_SOURCE = allocate_field1D(nz);
  NN_SOURCE = allocate_field2D(nz,nr);
  BASE_TKE_SOURCE = allocate_field1D(nz);
  BASE_BETA_SOURCE = allocate_field1D(nz);
  BASE_NN_SOURCE = allocate_field2D(nz,nr);
  WAVE_NN_SURFACE_FLUX = allocate_field1D(nr);
  set_field1D_to_value(WAVE_NN_SURFACE_FLUX,0.0);

  fepsilon = setup_fracwd(Tn, mean_epsilon, dt, T);

  /* setup the scale height for the plunging source */

  // Gemmerich & Farmer spilling source is 20cm so set
  // scale height to be 20cm/2 = 10cm

  //  scale_height = 0.1;
  
  scale_height = Hrms/2.0;

  /* init the tke & beta source */

  setup_tke_beta_sources2(BASE_TKE_SOURCE,BASE_BETA_SOURCE,scale_height);

  /* first calculate the basic spectrum shape */

  rd = Radius->data;
  nnsd = BASE_NN_SOURCE->data;
  M = Radius->N;
  N = nz;

  wnsfd = WAVE_NN_SURFACE_FLUX->data;

  for (i=0;i<M;i++) {
    rr = *rd++;
    aa = rr/a0;
    if (aa>1.0)
      tmp=pow(aa,-4.5);
    else {
      if (rr>2e-4) 
	tmp=pow(aa,-2.5);
      else
        tmp=0.0;
    }
    *nnsd++ = n0 * tmp;
    *wnsfd++ = tmp;
  }

  /* next calculate this void fraction */

  betat = calculate_void_fraction_point(BASE_NN_SOURCE, 0);
  betaf = calculate_void_fraction0(WAVE_NN_SURFACE_FLUX);

  mult_field1D_const(WAVE_NN_SURFACE_FLUX,1/betaf);  // normalized to one!

  /* then loop through all the depths to set N(z,r) source */

  for (j=1;j<N-1;j++) {
    beta = DR1(BASE_BETA_SOURCE,j);
    betam = beta / betat;
    nnsd = BASE_NN_SOURCE->data;
    nnsd1 = REF2(BASE_NN_SOURCE,j);
    for (i=0;i<M;i++) 
      *nnsd1++ = betam * (*nnsd++);
  }

  /*  update the boundary condition w/ no-flux at both ends */

  update_bubble_bc0(BASE_NN_SOURCE);  

#ifdef DEBUG  /* check that the BASE_NN_SOURCE is correct */
  for (i=1;i<N-1;i++) {
    betat = calculate_void_fraction_point(BASE_NN_SOURCE,i);
    betam = DR1(BASE_BETA_SOURCE,i);
    beta = fabs(betat-betam);
    if (beta > 1e-9) {
      fprintf(stdout,"** wave.c: init_wave_parameters(): At level %d, betat = %lf,  base_beta=%lf, diff = %e \n",i,betat, betam, beta);
    }
  }
#endif /* DEBUG */

  copy_field1D(TKE_SOURCE,BASE_TKE_SOURCE);
  copy_field2D(NN_SOURCE,BASE_NN_SOURCE);

  mult_field1D_const(TKE_SOURCE,0.5*fepsilon);
  mult_field2D_const(NN_SOURCE,0.5*fepsilon);
  mult_field1D_const(WAVE_NN_SURFACE_FLUX,0.5*fepsilon/grav);  // this is WRONG in units

}


double fraction_wave_dissipation(int time_index)
{

  double frac;
  static int time_index2=0;  
  static int local_index=0;

  frac = DR1(FRAC_WD,local_index);

  time_index2++;
  local_index++;
  if (local_index>=Tn)
    local_index = 0;

  return frac;

}

void calculate_wave_stats(int time_index, double dt)
{

  double frac_diss;
  double tot_tke_source, tot_bpe_source, tot_surface_bpe_source;
  double mfac1, mfac2, mfac3;
  static double old_tke_source=1.0;
  static double old_bpe_source=1.0;
  static double old_surface_bpe_source=1.0;
  
  double phase;
  double tt, time;

#ifdef WAVE_ASYMMETRY
  double asymmetry_factor = 0.0;
  double skewness_factor = 0.5;
#endif  /* WAVE_ASYMMETRY */

  /* update turbulent sources */
 
  wave_breaking = OFF;
  frac_diss = fraction_wave_dissipation(time_index);

  tot_tke_source = 0.5 * frac_diss;    //* mean wave epsilon;
  tot_bpe_source = 0.5 * frac_diss;    // * mean wave epsilon;
  tot_surface_bpe_source = 0.5*frac_diss/grav;  // !!!! this is WRONG!!

  /* setup the field1D sources */

  if (frac_diss>0) {

    switch (turb_source) {
    case PLUNGING:  wave_breaking=ON;
                    break;
    case SPILLING:  wave_breaking=ON;
                    wave_tke_surface_flux = 0.5*frac_diss;
		    break;
    case NONE:      wave_breaking=OFF;
    }

  } 
  else {
    wave_tke_surface_flux = 0.0;
  }


  if (wave_forcing==ON) {
    time = time_index*dt;
    tt = omega*time;
    phase = cos(tt);
#ifdef WAVE_ASYMMETRY
    phase = cos(tt) + skewness_factor*cos(2*tt) + asymmetry_factor*sin(2*tt);
#endif /* WAVE_ASYMMETRY */
    wave_forcx = max_forcx * phase;
    wave_forcy = max_forcy * phase;   
#ifdef SHAW_TIDAL_FORCING
    tt = omega_tide*time;
    wave_forcx += max_tidal_forc * cos(tt);
    wave_forcy += -max_tidal_forc * sin(tt);
#endif /* SHAW_TIDAL_FORCING */
  }
  else {
    wave_forcx=0.0;
    wave_forcy=0.0;
  }
}



void deallocate_wave_variables()
{

  deallocate_field1D(TKE_SOURCE);
  deallocate_field2D(NN_SOURCE);
  deallocate_field1D(BASE_TKE_SOURCE);
  deallocate_field2D(BASE_NN_SOURCE);
  deallocate_field1D(FRAC_WD);
  deallocate_field1D(WAVE_NN_SURFACE_FLUX);

}
